"""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.

Copyright (C) 2024 Gran Telescopio Canarias <https://www.gtc.iac.es>
Fabricio Manuel Pérez Toledo <fabricio.perez@gtc.iac.es>
"""

__author__="Fabricio Manuel Pérez Toledo"
__version__ = "0.2.0"
__license__ = "GPL v3.0"

from SAUSERO.check_files import *
from SAUSERO.reduction_osirisplus import *
from SAUSERO.aligning_osirisplus import *
from SAUSERO.astrometry_osirisplus import *
from SAUSERO.photometry_osirisplus import *

from astropy import units as u

import argparse, time, os, shutil
import os, json, warnings
import pkg_resources

from SAUSERO.Color_Codes import bcolors as bcl
from loguru import logger

warnings.filterwarnings("ignore", category=UserWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

# Parse configuration
#parser = argparse.ArgumentParser(
#                     prog = 'OsirisDRP',
#                     description = 'This software reduces observations taken with OSIRIS\
#                        in BBI mode. It can process any filter configuration and is suitable\
#                        for observations affected by fringing (Sloan_z).')
#
#parser.add_argument('-pr','--program', help='Select the GTC program.',
#                     required=True, type=str)
#
#parser.add_argument('-bl','--block', help='Select the block of the program.',
#                     required=True, type=str)
#
#args = parser.parse_args()

############## Predefined functions #############

def create_config_file_home():
    """
    This function creates a copy of the configuration file in .config/sausero/ for easier accessibility.
    """
    if not os.path.exists(f"{os.path.expanduser('~')}/sausero/configuration.json"):
        config_path = pkg_resources.resource_filename(
        'SAUSERO', 'config/configuration.json')
        os.makedirs(f"{os.path.expanduser('~')}/sausero/", exist_ok=True)
        shutil.copy(config_path,f'{os.path.expanduser("~")}/sausero/configuration.json')
        print(f"{bcl.WARNING}On the first run, it is necessary to relocate the configuration file to the ~/config/sausero/\n\
directory for easier accessibility. Navigate to this directory to modify the paths to the code location and\n\
the images to be processed. Once the modifications are made, you can start using SAUSERO.{bcl.ENDC}")
        sys.exit()


def readJSON():
    """
    Reads the file containing the configuration parameters.

    Returns:
        json: Collection of configuration parameters 
    """
    if os.path.exists(f"{os.path.expanduser('~')}/sausero/configuration.json"):
        return json.load(open(f"{os.path.expanduser('~')}/sausero/configuration.json"))
    else:
        config_path = pkg_resources.resource_filename(
            'SAUSERO', 'config/configuration.json')
        #return json.load(open("SAUSERO/config/configuration.json"))
        return json.load(open(config_path))
    

def Results(PATH, ZP, eZP, MASK, filt, ext_info = extinction_dict, conf = None):
    """ This function adds relevant information to the science image that will be 
    delivered to the PI.This information includes the instrumental magnitude for 
    the used filter and its error, the units used, the extinction value applied and 
    its error, dividing the image by the exposure time, and applying the mask.

    Args:
        PATH (string): Directory containing the preliminary results.
        ZP (float): The estimated instrumental magnitude for the filter.
        eZP (float):  The error of the instrumental magnitude.
        MASK (bool): Mask of bad pixels.
        filt (string): Filter used for the image acquisition.
    """
    ic = ccdp.ImageFileCollection(PATH, keywords='*', glob_include='*ast*')
    for sky in ['SKY', 'NOSKY']:
        if conf['REDUCTION']['save_not_sky'] or sky == 'SKY':
            fname = ic.files_filtered(include_path=True, filtro=filt, ssky=sky)[0]
            frame = CCDData.read(fname, unit='adu')
            logger.info(f'Load science image: {os.path.basename(fname)}')
            hd = frame.header
            hd['ZP'] = (ZP, 'ZeroPoint estimation')
            hd['eZP'] = (eZP, 'Error ZeroPoint estimation')
            logger.info('ZeroPoint information added to header')
            extinction, e_extinction = ext_info[filt]
            hd['EXT']=(extinction, 'Filter extinction')
            hd['eEXT']=(e_extinction, 'Error filter extinction')
            logger.info('Extinction information added to header')
            frame.header = hd
            frame.unit = u.adu/u.second
            logger.info('Change units: ADUs to ADUs/second')
            hd['PHOTOMETRY'] = (True, 'Photometry applied')

            frame.data = (frame.data / hd['EXPTIME'])
            frame.write(PATH / f"{hd['GTCPRGID']}_{hd['GTCOBID']}_{filt}_{(hd['DATE'].split('T')[0]).replace('-','')}_{sky}_BBI.fits",
                        overwrite=True)
            logger.info(f"Frame generated: {hd['GTCPRGID']}_{hd['GTCOBID']}_{filt}_{(hd['DATE'].split('T')[0]).replace('-','')}_{sky}_BBI.fits")
        else:
            logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The photometry is not going to be executed for NOSKY!!!!!!!{bcl.ENDC}')

def run():
    """
    This function 
    """
    # Parse configuration
    parser = argparse.ArgumentParser(
                         prog = 'OsirisDRP',
                         description = 'This software reduces observations taken with OSIRIS\
                            in BBI mode. It can process any filter configuration and is suitable\
                            for observations affected by fringing (Sloan_z).')

    parser.add_argument('-pr','--program', help='Select the GTC program.',
                         required=True, type=str)

    parser.add_argument('-bl','--block', help='Select the block of the program.',
                         required=True, type=str)

    args = parser.parse_args()


    print(f"{bcl.OKBLUE}***********************************************************************{bcl.ENDC}")
    print(f"{bcl.OKBLUE}************************* WELCOME TO SAUSERO **************************{bcl.ENDC}")
    print(f"{bcl.OKBLUE}***********************************************************************{bcl.ENDC}")
    print("\n")
    print(f"{bcl.BOLD}---------------------- LICENSE ----------------------{bcl.ENDC}")
    print("\n")
    print(f"This program is free software: you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation, either version 3 of the License, or\n\
(at your option) any later version.\n\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n\
GNU General Public License for more details.\n\n\
You should have received a copy of the GNU General Public License\n\
along with this program. If not, see <https://www.gnu.org/licenses/>.")
    print("\n")
    print(f"{bcl.BOLD}************************ IMPORTANT INFORMATION ************************{bcl.ENDC}")
    print("\n")
    print(f"This software is designed to reduce Broad Band Imaging observations obtained with OSIRIS+.\n\
For proper use, you need to modify the configuration file, which can be found\n\
in the directory where this software is installed. Additionally, you need to create\n\
an account on Astrometry.net. Once you have the code that allows you to use the API,\n\
you need to fill in the correct variable.")
    print(f"\n")

    create_config_file_home()

    PRG = args.program
    OB = args.block

    config_path = pkg_resources.resource_filename(
        'SAUSERO', 'config/configuration.json')
    check_files(config_path, PRG, OB)

    hora_local = time.localtime()
    conf = readJSON()

    logger.add(f"{conf['DIRECTORIES']['PATH_DATA']}{PRG}_{OB}/sausero_{time.strftime('%Y-%m-%d_%H:%M:%S', hora_local)}.log", format="{time} {level} {message} ({module}:{line})", level="INFO",
               filter=lambda record: 'astropy' not in record["name"])
    
    #Reduction Recipe. This recipe is responsible for cleaning the images by subtracting 
    #the masterbias and dividing by the normalized masterflat.
    #Subsequently, the cleaned images are saved.
    logger.info(f'{bcl.HEADER}---------- Starting the reduction ----------{bcl.ENDC}')
    logger.info("Configuration updated successfully.")
    bpm_path = pkg_resources.resource_filename(
        'SAUSERO', 'BPM/BPM_OSIRIS_PLUS.fits')
    o = Reduction(PRG, OB, main_path=conf['DIRECTORIES']['PATH_DATA'],
                path_mask=bpm_path)
    o.get_imagetypes()
    o.load_BPM()
    o.sort_down_drawer()

    if conf['REDUCTION']['use_BIAS']:
        o.do_masterbias()
    else:
        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The masterbias is not going to be created !!!!!!!{bcl.ENDC}')

    if conf['REDUCTION']['use_FLAT']:
        o.do_masterflat()
    else:
        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The masterflat is not going to be created !!!!!!!{bcl.ENDC}')
    
    if conf['REDUCTION']['use_STD']:
        o.get_std(no_CRs=conf['REDUCTION']['no_CRs'], contrast_arg = conf['REDUCTION']['contrast'],
            cr_threshold_arg = conf['REDUCTION']['cr_threshold'],
            neighbor_threshold_arg = conf['REDUCTION']['neighbor_threshold'], apply_flat=conf['REDUCTION']['use_FLAT'])
    else:
        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The STD star is not going to be reduced !!!!!!!{bcl.ENDC}')
    
    o.get_target(no_CRs=conf['REDUCTION']['no_CRs'], contrast_arg = conf['REDUCTION']['contrast'],
            cr_threshold_arg = conf['REDUCTION']['cr_threshold'],
            neighbor_threshold_arg = conf['REDUCTION']['neighbor_threshold'], apply_flat=conf['REDUCTION']['use_FLAT'])
    
    if conf['REDUCTION']['save_fringing']:
        o.remove_fringing()
    else:
        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The fringing correction is not going to be executed !!!!!!!{bcl.ENDC}')

    if conf['REDUCTION']['save_not_sky']:    
        o.sustract_sky()
    else:
        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The sky substraction is not going to be executed !!!!!!!{bcl.ENDC}')
    
    o.save_target(std=conf['REDUCTION']['save_std'])
    o.save_target(sky=conf['REDUCTION']['save_sky'])
    o.save_target(fringing=conf['REDUCTION']['save_fringing'])    
    o.save_target(not_sky=conf['REDUCTION']['save_not_sky'])
    logger.info(f'{bcl.HEADER}¡¡¡¡¡¡¡ Finished the reduction successfully !!!!!!!{bcl.ENDC}')
    print(2*"\n")
    
    #Aligned Recipe. The cleaned science images are aligned based on the filter used in each case. 
    #Then, they are saved as aligned images.
    if conf['ALIGNING']['use_aligning']:
        logger.info(f"{bcl.HEADER}---------- Starting the alignment ----------{bcl.ENDC}")
        al = OsirisAlign(PRG, OB, conf)
        for filt in list(set(al.ic.summary['filtro'])):
            for sky in ['SKY', 'NOSKY']:
                logger.info(f'{bcl.WARNING}++++++++++ Aligment for {filt} & {sky} ++++++++++{bcl.ENDC}')
                if conf['REDUCTION']['save_not_sky'] or sky == 'SKY':
                    align = al.aligning(filt, sky=sky)
                    lst = al.load_frames(filt, sky=sky)
                    fr = CCDData.read(lst[0], unit='adu')
                    header = fr.header
                    header['STACKED'] = (True, 'Stacked image')
                    header['exptime'] = al.total_exptime * (al.num + 1.)
                    logger.info(f"Estimated total exposure time: {header['exptime']} sec")
                    wcs = fr.wcs
                    logger.info(f"Updating the WCS information")
                    save_fits(align, header, wcs, al.PATH_REDUCED / f'{PRG}_{OB}_{filt}_stacked_{sky}.fits')
                    print(f'{bcl.HEADER}¡¡¡¡¡¡¡ Alineado para {filt} & {sky} realizado exitosamente !!!!!!!{bcl.ENDC}')
                else:
                    logger.warning(f'{bcl.WARNING}: Alignment is not going to be executed for NOSKY{bcl.ENDC}')

        print(f'{bcl.HEADER}¡¡¡¡¡¡¡ The alignment for each filter is finished !!!!!!!{bcl.ENDC}')
        print(2*"\n")
    else:
        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The alignment is not going to be executed !!!!!!!{bcl.ENDC}')

     #Astrometry Recipe. The aligned image for each filter undergoes an astrometric process to 
     #accurately determine the real positions of the celestial bodies present in the scene.
    if conf['ASTROMETRY']['use_astrometry']:
        logger.info(f"{bcl.HEADER}---------- Start the astrometrization ----------{bcl.ENDC}")
        del filt
        ic_ast = ccdp.ImageFileCollection(al.PATH_REDUCED, keywords='*', glob_include='*stacked*', glob_exclude='*NOSKY*')
        lst_filt = list(ic_ast.summary['filtro'])
        for filt in lst_filt:
            logger.info(f'{bcl.WARNING}++++++++++ Astrometrization for {filt} ++++++++++{bcl.ENDC}')
            best_wcs, new_frame = solving_astrometry(PRG, OB, filt, conf, sky='SKY', calib_std=False)
            for sky in ['SKY', 'NOSKY']:
                if sky == 'SKY':
                    new_frame.header['ASTROMETRY'] = (True, 'Astrometrized image')
                    new_frame.write(al.PATH_REDUCED / f'{PRG}_{OB}_{filt}_ast_SKY.fits', overwrite=True)
                else:
                    if conf['REDUCTION']['save_not_sky']:
                        nosky = CCDData.read(al.PATH_REDUCED / f'{PRG}_{OB}_{filt}_stacked_NOSKY.fits', unit='adu')
                        nosky.wcs = WCS(best_wcs)
                        nosky.header['ASTROMETRY'] = (True, 'Astrometrized image')
                        nosky.write(al.PATH_REDUCED / f'{PRG}_{OB}_{filt}_ast_NOSKY.fits', overwrite=True)
                    else:
                        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The astrometry is not going to be executed for NOSKY !!!!!!!{bcl.ENDC}')
                
                print(f'{bcl.HEADER}¡¡¡¡¡¡¡ Astrometrization done successfully for {filt} !!!!!!!{bcl.ENDC}')
                time.sleep(10)
    else:
        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The astrometry is not going to be executed !!!!!!!{bcl.ENDC}')

    if conf['REDUCTION']['use_STD'] and conf['ASTROMETRY']['use_astrometry']:
        logger.info(f'{bcl.HEADER}---------- Start astrometrization for STD star ----------{bcl.ENDC}')
        time.sleep(30)
        ic_std = ccdp.ImageFileCollection(al.PATH_REDUCED, keywords='*', glob_include='*STD*', glob_exclude='*NOSKY*')
        best_wcs_std, new_frame_std = solving_astrometry(PRG, OB, filt, conf, sky='SKY', calib_std=True)
        for path_to_std in ic_std.files_filtered(include_path=True):
            logger.info(f'{bcl.WARNING}++++++++++ Astrometrization for: {path_to_std} ++++++++++{bcl.ENDC}')
            std_img = CCDData.read(path_to_std, unit='adu')
            std_img.wcs = WCS(best_wcs_std)
            std_img.header['ASTROMETRY'] = (True, 'Astrometrized image')
            std_img.write(path_to_std, overwrite=True)


        print(f'{bcl.HEADER}¡¡¡¡¡¡¡ Astrometrization finished !!!!!!!{bcl.ENDC}')
        print(2*"\n")
    else:
        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The astrometry for STD is not going to be executed !!!!!!!{bcl.ENDC}')


    #Photometry Recipe. The instrumental magnitude is estimated from the calibration star. This allows us 
    # to later estimate the apparent magnitude of the celestial bodies present in the science image. 
    # This process is done for each filter used. Additionally, the Results() function allows including 
    # this information in the header of the cleaned, aligned, and astrometrically processed science image.
    if conf['PHOTOMETRY']['use_photometry']:
        logger.info(f"{bcl.HEADER}---------- Starting the estimation of ZeroPoint ----------{bcl.ENDC}")
        del filt
        ic_pho = ccdp.ImageFileCollection(al.PATH_REDUCED, keywords='*', glob_include='*red*')
        lst_filt = list(set(ic_pho.summary['filtro']))
        for filt in lst_filt:
            logger.info(f'{bcl.WARNING}++++++++++ Filter selected is {filt} ++++++++++{bcl.ENDC}')
            ZP, eZP = photometry(PRG, OB, ic_pho.files_filtered(imgtype='STD',filter2=filt)[0], conf)
            Results(al.PATH_REDUCED, ZP, eZP, o.MASK, filt, conf = conf)

        logger.info(f'{bcl.HEADER}¡¡¡¡¡¡¡ End of the reduction. The results are available in reduced directory. !!!!!!!{bcl.ENDC}')
    else:
        logger.warning(f'{bcl.WARNING}¡¡¡¡¡¡¡ The photometry is not going to be executed !!!!!!!{bcl.ENDC}')

if __name__ == '__main__':
    run()



"""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.

Copyright (C) 2024 Gran Telescopio Canarias <https://www.gtc.iac.es>
Fabricio Manuel Pérez Toledo <fabricio.perez@gtc.iac.es>
"""

class bcolors:
    """
    A class to define string color formatting for terminal output using ANSI escape sequences.

    This class provides various color codes and text styles that can be used to format terminal output. 
    The colors and styles are defined as class variables, and each variable represents a different format 
    that can be applied to text in the terminal.

    Class variables:
        HEADER (str): Color for header text.
        OKBLUE (str): Color for informational text (blue).
        OKCYAN (str): Color for informational text (cyan).
        OKGREEN (str): Color for successful operation text (green).
        WARNING (str): Color for warning text (yellow).
        FAIL (str): Color for error or failed operation text (red).
        ENDC (str): Resets the text formatting to default.
        BOLD (str): Bold text style.
        UNDERLINE (str): Underlined text style.
    """
    HEADER = '\033[33m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


"""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.

Copyright (C) 2024 Gran Telescopio Canarias <https://www.gtc.iac.es>
Fabricio Manuel Pérez Toledo <fabricio.perez@gtc.iac.es>
"""

import os, sys, time, json, logging
from pathlib import Path

from astropy import units as u
from astropy import wcs
from astropy.nddata import CCDData
from astropy.io import fits
import ccdproc as ccdp
from matplotlib import pyplot as plt
import numpy as np
import yaml as py
import lacosmic
import sep

from SAUSERO.Color_Codes import bcolors as bcl
from loguru import logger

#logger = logging.getLogger("main_logger")

import logging, inspect
class InterceptHandler(logging.Handler):
    """Custom logging handler that redirects log records from the standard logging module 
    to the Loguru logger.

    This handler allows the use of Loguru alongside the standard logging module, ensuring 
    that logs captured by the standard logging system are properly forwarded to the Loguru logger.

    Args:
        logging (module): The logging module that provides the log records to be captured 
                           and forwarded to Loguru. Typically, this would be the standard 
                           `logging` module.
    """
    
    def emit(self, record: logging.LogRecord) -> None:
        """
        Process a log record and forward it to the Loguru logger, mapping the log level and 
        determining the caller's origin for the log message.

        This method retrieves the log level, the caller's information, and the log message 
        before passing it to Loguru for proper logging.

        Args:
            record (logging.LogRecord): The log record to be processed, which contains 
                                        information about the log level, message, and other 
                                        relevant details for the log entry.

        This method will convert the logging level to a corresponding Loguru level and 
        propagate the message along with the exception information (if any) to the Loguru logger.
        """
        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        if level != "DEBUG":
            logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())

logging.basicConfig(handlers=[InterceptHandler()], level=logging.INFO, force=False)

class Reduction:
    """The goal is to perform the cleaning procedure for science and photometric calibration frames. 
    First, bias frames are averaged to create a master bias. This master bias is then used to subtract 
    the offset level from the sky flat. The sky flats are subsequently combined and divided by their 
    median to normalize them. The final product is a master flat, which will be used to standardize 
    the sensor's response. Finally, the science and photometric calibration frames are corrected by 
    subtracting the master bias and dividing by the master flat, resulting in the final reduced frames.
    """

    def __init__(self, gtcprgid, gtcobid, main_path, path_mask = None):
        """Object initialization to carry out the reduction.

        Args:
            gtcprgid (str): Observation program code.
            gtcobid (str): Observation block code.
            path_mask (str, optional): Path to BPM.
        """
        logger.info(f'Initializing the reduction for {gtcprgid}_{gtcobid}.')
        
        self.gtcprgid = gtcprgid
        self.gtcobid = gtcobid
        ROOT = main_path

        # The path containing the raw images is defined
        self.PATH = Path(ROOT + str(gtcprgid) + '_' + str(gtcobid) + '/' + 'raw/')
        if os.path.exists(self.PATH):
            logger.info("Path to raw data exists")
        else:
            logger.critical(f"Path to raw data does NOT exist")
            sys.exit()

        # The directory that will contain the intermediate and processed images is defined
        self.PATH_RESULTS = Path(ROOT + str(gtcprgid) + '_' + str(gtcobid) + '/' + 'reduced/')
        self.PATH_RESULTS.mkdir(parents=True, exist_ok=True)
        if os.path.exists(self.PATH_RESULTS):
            logger.info("Path to reduced directory has been created")
        else:
            logger.critical(f"Path to reduced directory has NOT been created")
            sys.exit()

        # Define the path to mask file
        self.path_mask = Path(path_mask)
        if os.path.exists(self.path_mask):
            logger.info("Path to mask file exists")
        else:
            logger.critical(f"Path to mask file does NOT exist")
            sys.exit()
        
        # The information about the frames in that directory is gathered.
        self.ic = ccdp.ImageFileCollection(self.PATH)
        if len(self.ic.summary) != 0:
            logger.info("Data collection is ready")
        else:
            logger.critical(f"NOT files to reduce")
            sys.exit()

        # Define dictionaries
        self.DATA_DICT={}
        self.key_dict={'flat':'OsirisSkyFlat',
                       'target': 'OsirisBroadBandImage',
                       'std': 'OsirisBroadBandImage',
                       'bias': 'OsirisBias'
        }
        self.master_dict = {} #Ready to use
        self.std_dict = {}
        self.target_dict = {}

        logger.info("Dictionaries have been created")




    @staticmethod
    def configure_mask(mask):
        """Static method to reshape the BPM array to match any frame shape.

        Args:
            mask (int): BPM HDU or CCDData Object

        Returns:
            bool: BPM array in boolean format to apply over the frames.
        """
        mask = mask.data.astype(bool)
        matrix = np.ones(mask.shape)
        matrix[mask == False] = np.nan
        logger.info("BPM is ready.")
        return matrix[230:2026,28:2060] # TRIM SECTION




    @staticmethod
    def get_each_data(data_dict, value):
        """Reads the contents of a dictionary containing the paths to a series of 
        frames and opens them, adding them to a list.

        Args:
            data_dict (dict): A dictionary containing the paths to 
            one or more frames in a list.
            value (str): Key to access the list of paths.

        Returns:
            list: It is a list containing the images.
        """
        ccd = []
        for frame_path in data_dict[value]:
            hdul = fits.open(frame_path)
            ccd.append(hdul[0].data[230:2026,28:2060]) #TRIM SECTION
            
        logger.info(f"List of images for key {value} is ready.")
        return ccd



    @staticmethod
    def combining(lst_frames):
        """This static method combines the images in a list to obtain an averaged image. 
        The process involves creating a data cube and averaging the images.

        Args:
            lst_frames (list): List of images to be averaged.

        Returns:
            image: Create an averaged matrix from a data cube.
        """
        cube = np.dstack(lst_frames)
        cube.sort(axis=2)
        return np.nanmedian(cube[:,:,1:-1], axis=2)




    def create_cubes(self, key):
        """Create a data cube from a list of images.

        Args:
            key (str): Key of the scientific images dictionary.

        Returns:
            cube(float): Data cube.
        """
        return np.dstack(self.target_dict[key],
                        axis=2)




    def sustractMasterBias(self, value, master, data_dict):
        """Subtract the master bias from the image.

        Args:
            value (str): Key that allows access to the images in 
            the dictionary.
            master (float): MasterBias frame.
            data_dict (dict): Images dictionary. 

        Returns:
            list: List of the frames with masterbias applied.
        """
        ccd = self.get_each_data(data_dict, value)

        frames = [fr - master for fr in ccd]

        logger.info(f"Masterbias applied to key {value} set.")
        
        return frames




    def clean_target(self, value, masterbias, masterflat):
        """Applies the subtraction of the MasterBias and the division 
        by the normalized MasterFlat to the science images.

        Returns:
            list: List of cleaned science frames.
        """
        #value, masterbias, masterflat = args
        ccd = self.get_each_data(self.DATA_DICT, value)
        if masterflat is None:
            masterflat = 1.0
        frames = [(fr - masterbias)/masterflat for fr in ccd]
        logger.info(f"Applied masterbias and masterflat on the frames for {value}.")
        return frames


    def get_imagetypes(self):
        """ This method aims to identify all the types of images present in the original images 
        directory and their corresponding filters. To achieve this, it creates an empty dictionary 
        where each key is a combination of the image type and the filter used.
        """
        logger.info('Getting types of images and filters used.')
        self.filt_wheels = []
        matches = (self.ic.summary['obsmode'] != 'OsirisBias')
        matches1 = (self.ic.summary['obsmode'] != 'OsirisBias') & (self.ic.summary['filter1'] != 'OPEN')
        matches2 = (self.ic.summary['obsmode'] != 'OsirisBias') & (self.ic.summary['filter2'] != 'OPEN')
        matches3 = (self.ic.summary['obsmode'] != 'OsirisBias') & (self.ic.summary['filter3'] != 'OPEN')

        if len(list(set(self.ic.summary['filter1'][matches1]))) >= 1:
            self.filt_wheels.append('filter1')
        elif len(list(set(self.ic.summary['filter2'][matches2]))) >= 1:
            self.filt_wheels.append('filter2')
        elif len(list(set(self.ic.summary['filter3'][matches3]))) >= 1:
            self.filt_wheels.append('filter3')
        else:
            raise(ValueError, "All filters are open!!!")
            sys.exit()

        for filt in self.filt_wheels:
            for value in list(set(self.ic.summary[filt][matches])):
                for type in ['flat', 'std', 'target']:
                    self.DATA_DICT[type + '+' + value] = []

        self.DATA_DICT["bias"] = []


    def load_BPM(self):
        """
        This method opens the FITS file containing the BPM.
        """
        bpm = CCDData.read(self.path_mask, unit=u.dimensionless_unscaled,
                            hdu=0, format='fits', ignore_missing_simple=True)
        
        #bpm = np.ones((2056,2073))

        self.MASK = self.configure_mask(bpm)
        logger.info("BPM is ready.")




    def load_results(self):
        """It generates a table with specific content in the results directory, specifically for 
        those scientific images that have already been cleaned.
        """
        self.ic_r = ccdp.ImageFileCollection(self.PATH_RESULTS, keywords='*',
                                             glob_include='*red*') # He modificado el glob_include
        logger.info("Table with several reduced science frames is ready.")




    def sort_down_drawer(self):
        """This method is responsible for storing the images by their type and filter in the previously created dictionary.
        """
        for filt in self.filt_wheels:
            for elem in list(self.DATA_DICT.keys()):
                if elem != 'bias':
                    key, value = elem.split('+')
                    types_targets = set(self.ic.summary['object'][self.ic.summary['obsmode'] == 'OsirisBroadBandImage'])
                    if key == 'flat':
                        tmp_dict = {"obsmode":self.key_dict[key], filt: value}
                        logger.info("Including flat frames.")
                    elif key == 'std':
                        target_type = [data for data in types_targets if 'STD' in data][0]
                        tmp_dict = {"obsmode":self.key_dict[key], filt: value, 'object':target_type}
                        logger.info("Including photometric calibration frames.")
                    elif key == 'target':
                        target_type = [data for data in types_targets if not 'STD' in data][0]
                        tmp_dict = {"obsmode":self.key_dict[key], filt: value, 'object':target_type}
                        logger.info("Including science frames.")
                    self.DATA_DICT[elem] = self.ic.files_filtered(**tmp_dict, include_path=True)
                else:
                    self.DATA_DICT[elem] = self.ic.files_filtered(obsmode='OsirisBias', include_path=True)
                    logger.info("Including bias frames.")



    def do_masterbias(self):
        """
        This method creates the master bias frame.
        """
        bias = self.get_each_data(self.DATA_DICT, "bias")

        self.masterbias = self.combining(bias) * self.MASK
        self.master_dict['bias'] = self.masterbias

        logger.info("Masterbias has been created.")



    def do_masterflat(self):
        """
        This method creates the master flat frame for each filter.
        """
        lst_flat = [elem for elem in list(self.DATA_DICT.keys()) if 'flat' in elem]
        for filt in lst_flat:
            flat = self.sustractMasterBias(filt, self.masterbias, self.DATA_DICT)
            combflat = self.combining(flat)
            median = np.nanmedian(combflat)
            masterflat = combflat/median
            self.master_dict[filt] = masterflat
            logger.info(f"Masterflat has been created for {filt} filter.")


    def get_std(self, no_CRs=False, contrast_arg = 1.5, cr_threshold_arg = 5.,
                neighbor_threshold_arg = 5., apply_flat=False):
        """
        This method processes the photometric calibration frames.
        """
        lst_std = [elem for elem in list(self.DATA_DICT.keys()) if 'std' in elem]
        logger.info("Processing photometric calibration frames.")
        for elem in lst_std:
            key, value = elem.split('+')
            
            if apply_flat == False:
                self.master_dict['flat+' + value] = None

            #args = [elem, self.master_dict['bias'],
            #        self.master_dict['flat+' + value]]
            
            std = self.clean_target(elem, self.master_dict['bias'],self.master_dict['flat+' + value])
            lst_sd = []
            for sd in std:
                if no_CRs:
                    logger.info(f"Removing CRs to photometric calibration frame for {value}.")
                    no_mask = np.nan_to_num(sd, nan=np.nanmedian(sd))
                    lst_sd.append(lacosmic.lacosmic(no_mask, contrast=contrast_arg,
                                                    cr_threshold=cr_threshold_arg,
                                                    neighbor_threshold=neighbor_threshold_arg,
                                                    effective_gain=1.9,
                                                    readnoise=4.3)[0])
                else:
                    logger.info(f"NOT treatment for CRs applied to photometric calibration frame for {value}.")
                    lst_sd.append(np.nan_to_num(sd, nan=np.nanmedian(sd)))
            self.std_dict[elem] = lst_sd



    def get_target(self, no_CRs=False, contrast_arg = 1.5, cr_threshold_arg = 5.,
                neighbor_threshold_arg = 5., apply_flat = False):
        """
        This method cleans the science frames.
        """
        lst_target = [elem for elem in list(self.DATA_DICT.keys()) if 'target' in elem]
        logger.info("Processing science frames.")
        for elem in lst_target:
            key, value = elem.split('+')
            
            if apply_flat == False:
                self.master_dict['flat+' + value] = None
            
            #args = [elem, self.master_dict['bias'],
            #        self.master_dict['flat+' + value]]
             
            target= self.clean_target(elem, self.master_dict['bias'],self.master_dict['flat+' + value])
            lst_tg = []
            for tg in target:
                if no_CRs:
                    logger.info(f"Removing CRs to science frames for {value}.")
                    no_mask = np.nan_to_num(tg, nan=np.nanmedian(tg))
                    lst_tg.append(lacosmic.lacosmic(no_mask, contrast=contrast_arg,
                                                    cr_threshold=cr_threshold_arg,
                                                    neighbor_threshold=neighbor_threshold_arg,
                                                    effective_gain=1.9,
                                                    readnoise=4.3)[0])
                else:
                    logger.info(f"NOT treatment for CRs applied to science frames for {value}.")
                    lst_tg.append(np.nan_to_num(tg, nan=np.nanmedian(tg)))
            self.target_dict[elem] = lst_tg



    def remove_fringing(self):
        """
        This method performs a special cleaning when using Sloan_z to remove the interference pattern.
        """
        lst_results = [elem for elem in list(self.DATA_DICT.keys())]
        if 'Sloan_z' in lst_results:
            logger.info("Removing the fringe on Sloan z filter.")
            fringe= self.target_dict['target+Sloan_z']
            combfringe = self.combining(fringe)
            median = np.nanmedian(combfringe)
            masterfringe = combfringe/median
            fr_free = [elem/masterfringe for elem in fringe]
            self.target_dict['fringe+Sloan_z'] = fr_free
            logger.info("Frames with fringe free.")

        
    
    def sustract_sky(self):
        """
        This method subtracts the contribution of the sky background.
        """
        lst_target_keys = [elem for elem in list(self.target_dict.keys())]
        logger.info(f"Substracting sky background.")
        for elem in lst_target_keys:
            key, value = elem.split('+')
            lst_frames = self.target_dict['target+' + value]
            cube = np.dstack(lst_frames)
            cube.sort(axis=2)
            im_avg = np.median(cube[:,:,:], axis=2)
            logger.info(f"Creating sky background simulated for {value}.")
            self.bkg = sep.Background(im_avg)
            no_sky = []
            for fr in lst_frames:
                no_sky.append(fr-self.bkg)
            self.target_dict['sky+' + value] = no_sky
            if key == 'STD':
                logger.info(f"List of photometric calibration frames without sky for {value} created.")
            elif key == 'target':
                logger.info(f"List of science frames without sky for {value} created.")
            else:
                logger.error("No defined option for key.")
            


    def save_target(self, fringing=False, std=False, sky=False, not_sky=False):
        """This method saves the images generated during the cleaning process and 
        adds information to the header to assist in future processes.

        Args:
            fringing (bool, optional): Indicates whether the interference 
            pattern correction has been performed. Defaults to False.
            std (bool, optional): Indicates whether the image(s) to be
            saved are photometric calibration star images. Defaults to False.
            sky (bool, optional): Indicates whether the science image(s) to 
            be saved contain sky or not. Defaults to False.
        """
        if not std:
            logger.info("Saving reduced photometric calibration frames.")
            lst_results = [elem for elem in list(self.DATA_DICT.keys()) if 'target' in elem]
        else:
            logger.info("Saving science frames reduced.")
            lst_results = [elem for elem in list(self.DATA_DICT.keys()) if 'std' in elem]

        for key in lst_results:
            fnames = self.DATA_DICT[key]

            if key == 'target+Sloan_z' and fringing:
                FRIGING = 'NO'
                sky_status = 'SKY'
                status='REDUCED'
                imagetype='SCIENCE'
                target = self.target_dict['fringe+Sloan_z']
                logger.info("Science frames without fringe.")
            elif std:
                FRIGING = 'YES'
                sky_status = 'SKY'
                status='REDUCED'
                imagetype='STD'
                filt = key.split('+')[1]
                target= self.std_dict[key]
            elif sky:
                FRIGING = 'YES'
                sky_status = 'SKY'
                status='REDUCED'
                imagetype='SCIENCE'
                filt = key.split('+')[1]
                target= self.target_dict[key]
                logger.info("Science frames WITH sky.")
            elif not_sky:
                FRIGING = 'YES'
                sky_status = 'NOSKY'
                status='REDUCED'
                imagetype='SCIENCE'
                filt = key.split('+')[1]
                target= self.target_dict['sky+'+ filt]
                logger.info("Science frames WITHOUT sky.")
            else:
                logger.warning("No defined option for saving the frames.")
                break

            for i in range(len(fnames)):
                t = time.gmtime()
                time_string = time.strftime("%Y-%m-%dT%H:%M:%S", t)
                hd = fits.open(fnames[i])[0].header
                hd_wcs = wcs.WCS(fits.open(fnames[i])[0].header).to_header()
                hd['imgtype'] = imagetype
                hd['STATUS'] = status
                hd['SSKY'] = sky_status
                hd['BPMNAME'] = 'BPM_5sig' #Tipo de BPM aplicada.
                hd['rdate'] = time_string #Fecha de la reducción
                hd['filtro'] = filt #Simplifica la búsqueda por filtro.
                primary_hdu = fits.PrimaryHDU(target[i], header=hd+hd_wcs)
                hdul = fits.HDUList([primary_hdu])

                filename = os.path.basename(fnames[i])
                raw_name , __ = filename.split('.')
                #logger.info(f"Storing the frame: reduced_5sig_{adjetive}_{sky_status}_{filename} for {hd['FILTER2']}")
                #hdul.writeto(str(self.PATH_RESULTS / (f'reduced_5sig_{adjetive}_{sky_status}_' + filename)),
                #            overwrite=True)

                logger.info(f"Storing the frame: ADP_{imagetype}_{sky_status}_{raw_name} for {hd['FILTER2']}")
                hdul.writeto(str(self.PATH_RESULTS / (f'ADP_{imagetype}_{sky_status}_{raw_name}_red.fits')),
                            overwrite=True)


"""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.

Copyright (C) 2024 Gran Telescopio Canarias <https://www.gtc.iac.es>
Fabricio Manuel Pérez Toledo <fabricio.perez@gtc.iac.es>
"""

import yaml, glob, os

from astropy.nddata import CCDData
from astropy.wcs import WCS
from astrometry_net_client import Session, FileUpload, Settings

from SAUSERO.Color_Codes import bcolors as bcl
from loguru import logger

def settings(PATH_TO_CONFIG_FILE):
    """ This method opens the configuration file containing the parameters 
    needed for optimal astrometrization.

    Args:
        PATH_TO_CONFIG_FILE (str): Path to configuration file.

    Returns:
        dict: Contents of the configuration file.
    """
    try:
        print('Loading settings for astrometrization...')
        with open(PATH_TO_CONFIG_FILE, 'r') as file:
            prime_service = yaml.safe_load(file)
    except FileNotFoundError:
        raise('Not found the configuration file.')

    return prime_service['OSIRIS']

def apply_astrometrynet_client(filename, conf):
    """This method sends the stacked science frame along with settings information 
    to the Astrometry.net server through its API. Then, it receives the 
    WCS estimated by the server.

    Args:
        filename (str): Name of stacked science frame
        conf (dict): Collection of setting parameters

    Returns:
        WCS: WCS information given by the server.
    """    
    logger.info("Loading the settings")
    ss = Settings()
    img = CCDData.read(filename, unit='adu')
    #ss.scale_units = 'arcsecperpix'
    ss.set_scale_estimate(conf["ASTROMETRY"]["set_scale_estimate"]["scale"],
                        conf["ASTROMETRY"]["set_scale_estimate"]["unknown"], 
                        unit=conf["ASTROMETRY"]["set_scale_estimate"]["scale_units"])
    ss.center_ra = img.header['RADEG']
    ss.center_dec = img.header['DECDEG']
    ss.radius = conf["ASTROMETRY"]["radius"]
    ss.downsample_factor = conf["ASTROMETRY"]["downsample_factor"]
    ss.use_sextractor = conf["ASTROMETRY"]["use_sextractor"]
    ss.crpix_center = conf["ASTROMETRY"]["crpix_center"]
    ss.parity = conf["ASTROMETRY"]["parity"]
    ss.allow_commercial_use = 'n'
    ss.allow_modifications = 'n'
    ss.publicly_visible = 'n'
    logger.info("Settings:")
    logger.info(f"{ss}")
    #Send the image
    s = Session(api_key=conf["ASTROMETRY"]["No_Session"])
    logger.info("API connection is ready")
    upl = FileUpload(filename, session=s, settings=ss)
    logger.info("Frame has been uploaded")
    submission = upl.submit()
    logger.info("Waiting for an answer from API...")
    submission.until_done()
    job = submission.jobs[0]
    job.until_done()
    if job.success():
        wcs = job.wcs_file()
        logger.info("WCS received from API")
    logger.info(job.info())
    
    if wcs != None:
        return wcs

def modify_WCS(best_wcs, PATH_TO_FILE):
    """This method updates the original WCS with the WCS estimated by the Astrometry.net server.

    Args:
        best_wcs (str): WCS estimated by the Astrometry.net server
        PATH_TO_FILE (str): Path to science frame

    Returns:
        CCDData: Science frame with its WCS updated.
    """
    frame = CCDData.read(PATH_TO_FILE, unit='adu')
    best_wcs = WCS(best_wcs)
    new_frame =  CCDData(data=frame.data, header=frame.header, wcs=best_wcs,
                         unit='adu')
    new_frame.write(PATH_TO_FILE, overwrite=True)
    logger.info(f"The WCS for {os.path.basename(PATH_TO_FILE)} has been updated")
    return new_frame

def solving_astrometry(PRG, OB, filt, conf, sky, calib_std = False):
    """This method handles the procedure for obtaining the astrometry solution from the server.

    Args:
        PRG (str): Science program code
        OB (str): Observational block number assigned to a science program
        filt (str): Filter name

    Returns:
        WCS, list, CCDData: Several results: the new WCS, a local catalogue for 
        the FoV using the new WCS to estimate positions, and the science frame 
        with its WCS updated.
    """
    root_path = conf["DIRECTORIES"]["PATH_DATA"]

    if calib_std:
        logger.info("Astrometry calibration for STD star")
        LST_PATH_TO_FILE = glob.glob(root_path + PRG + '_' + OB + '/reduced/' + f'*STD*.fits')
    else:
        logger.info("Astrometry calibration for science target")
        LST_PATH_TO_FILE = [root_path + PRG + '_' + OB + '/reduced/' + f'{PRG}_{OB}_{filt}_stacked_{sky}.fits']

    PATH_TO_FILE = LST_PATH_TO_FILE[0]

    best_wcs = apply_astrometrynet_client(PATH_TO_FILE, conf)

    new_frame = modify_WCS(best_wcs, PATH_TO_FILE)

    return best_wcs, new_frame #pixels



"""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.

Copyright (C) 2024 Gran Telescopio Canarias <https://www.gtc.iac.es>
Fabricio Manuel Pérez Toledo <fabricio.perez@gtc.iac.es>
"""

from astropy.nddata import CCDData
import numpy as np
import matplotlib.pyplot as plt
import sep
from astroquery.simbad import Simbad
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.coordinates import FK5
from astropy.visualization import LogStretch,imshow_norm, ZScaleInterval
from matplotlib.patches import Ellipse

from SAUSERO.Color_Codes import bcolors as bcl
from loguru import logger

extinction_dict = {
    'Sloan_u': [0.45, 0.02],
    'Sloan_g': [0.15, 0.02],
    'Sloan_r': [0.07, 0.01],
    'Sloan_i': [0.04, 0.01],
    'Sloan_z': [0.03, 0.01]
}

def get_ugriz(values):
    """From Sloan r apparent magnitude and a small list of the color,
    this function estimates the magnitude for each Sloan band.

    Args:
        values (list): List with Sloan r magnitude and colors. The order
        must be Sloan r, u-g , g-r, r-i, i-z. 

    Returns:
        float: Five magnitudes: Sloan u, Sloan g, Sloan r, Sloan i, Sloan z. 
    """
    r, ug, gr, ri, iz = values
    g = gr + r
    u = ug + g
    i = -ri + r
    z = -iz + i
    return u, g, r, i, z

def photometry(programa, bloque, filename, conf):
    """This method estimates the instrumental magnitude (zeropoint) for the night, depending on the filter used.

    Args:
        programa (str): Science program code
        bloque (str): Observational block number assigned to a science program
        filename (str): Name of STD star file.

    Returns:
        float: Estimation of the instrumental magnitude and its error.
    """
    # Dict with STD star collection: Coordinates, Sloan r magnitude and colors.
    bands_dict = {
        'G158-100': ['00 33 54.60', '-12 07 58.9', 14.691, 1.101, 0.510, 0.222, 0.092],
        'SA92-282': ['00 56 46.86',	'+00 38 30.9', 12.936, 1.000, 0.136, 0.021, -0.009],
        'Feige22': ['02 30 16.62',	'+05 15 50.6', 13.024, 0.050, -0.333, -0.303, -0.273],
        'SA95-193': ['03 53 20.59' , '+00 16 34.7', 13.844, 2.489, 1.097, 0.407, 0.214],
        'Ross49': ['05 44 56.81','+09 14 32.1',11.163, 1.130, 0.467, 0.162, 0.049],
        'Hilt_566': ['06 32 09.67','+03 34 44.3',10.787, 1.125, 0.673, 0.341, 0.211],
        'Ru-149F': ['07 24 14.02','-00 31 38.2', 13.119, 2.469, 0.867, 0.317, 0.166],
        'SA100-280': ['08 53 34.9','-00 37 38.5',11.689, 1.143, 0.308, 0.084, 0.003],
        'PG0918+029D': ['09 21 34.0','+02 46 39.0',11.937, 2.227, 0.817, 0.324, 0.166],
        'SA101-316': ['09 54 52.03','-00 18 34.5',11.438, 1.152, 0.309, 0.073, 0.007],
        'G162-66': ['10 33 42.81','-11 41 38.7', 13.227, -0.183, -0.387, -0.354, -0.303],
        'Feige_34': ['10 39 36.74','+43 06 09.2',11.423, -0.509, -0.508, -0.347, -0.265],
        'PG1047+003A': ['10 50 09.0','-00 01 08.0',13.303, 1.385, 0.519, 0.212, 0.087],
        'G163-50': ['11 07 59.97','-05 09 26.0', 13.266, 0.215, -0.277, -0.272, -0.271],
        'Feige66': ['12 37 23.52','+25 03 59.9',10.747, -0.345, -0.476, -0.367, -0.316],
        'SA104-428': ['12 41 41.29','-00 26 26.1',12.330, 2.153, 0.763, 0.279, 0.147],
        'PG1323-086D': ['13 26 05.26', '-08 50 35.7', 11.928, 1.210, 0.397, 0.132, 0.032],
        'Ross_838': ['14 01 44.48','+08 55 17.5',11.327, 1.277, 0.573, 0.239, 0.111],
        'PG1528+062B': ['15 30 39.55', '+06 01 13.1', 11.828, 1.235, 0.419, 0.143, 0.036],
        'G15-24': ['15 30 41.76','+08 23 40.4', 11.277, 1.035, 0.412, 0.151, 0.052],
        'BD+33 2642': ['15 51 59.89','+32 56 54.3',10.979, -0.018, -0.332, -0.284, -0.212],
        'Ross_530': ['16 19 51.65','+22 38 20.4',11.319, 1.273, 0.558, 0.229, 0.103],
        'Wolf_629': ['16 55 25.22','-08 19 21.3',11.129, 3.013, 1.413, 1.466, 0.648],
        'SA109-381': ['17 44 12.27','-00 20 32.8',11.514, 1.477, 0.547, 0.223, 0.094],
        'Ross_711': ['18 35 19.17','+28 41 55.4',11.295, 0.837, 0.282, 0.104, 0.015],
        'SA110-232': ['18 40 52.33', '+00 01 54.8', 12.287, 1.390, 0.552, 0.237, 0.094],
        'SA111-1925': ['19 37 28.64','+00 25 02.8',12.345, 1.397, 0.200, 0.061, 0.051],
        'Wolf_1346': ['20 34 21.88','+25 03 49.8',11.753, -0.016, -0.351, -0.309, -0.291],
        'SA112-805': ['20 42 46.75','+00 16 08.1',12.174, 1.183, -0.087, -0.135, -0.090],
        'BD+28 4211': ['21 51 11.02','+28 51 50.4',10.750, -0.517, -0.511, -0.379, -0.313],
        'G93-48': ['21 52 25.37','+02 23 19.6', 12.96, 10.107, -0.308, -0.307, -0.261],
        'SA114-656': ['22 41 35.07','+01 11 09.8',12.326, 1.961, 0.756, 0.293, 0.156],
        'GD_246': ['23 12 21.62','+10 47 04.3',13.346, -0.491, -0.504, -0.378, -0.367],
        'PG2336+004B': ['23 38 38.26','+00 42 46.4', 12.312, 1.101, 0.336, 0.100, 0.014]
    }
    
    # Extintion coefficients depends on the filter used.
    extinction_dict = {
        'Sloan_u': [0.45, 0.02],
        'Sloan_g': [0.15, 0.02],
        'Sloan_r': [0.07, 0.01],
        'Sloan_i': [0.04, 0.01],
        'Sloan_z': [0.03, 0.01]
    }
    
    #index_dict = {
    #    'Sloan_u': 0,
    #    'Sloan_g': 1,
    #    'Sloan_r': 2,
    #    'Sloan_i': 3,
    #    'Sloan_z': 4
    #}
    
    path = conf["DIRECTORIES"]["PATH_DATA"] + f"{programa}_{bloque}/reduced/"

    frame = CCDData.read(path + filename, unit='adu')
    logger.info("STD frame has been loaded successfully")

    hd = frame.header
    W = frame.wcs

    filtro = hd['FILTER2']
    #ind = index_dict[filtro]
    t = hd['EXPTIME']
    target_name = hd['OBJECT'].split('_')[1]
    
    ra, dec = bands_dict[target_name][:2]
    logger.info("Giving the coordinates for STD")

    c = SkyCoord(ra,dec, frame=FK5, unit=(u.hourangle, u.deg), obstime="J2000")

    x,y = W.world_to_pixel(c)
    logger.info("Transformed the original coordinates to pixel on the image.")

    fig = plt.figure(figsize=(15,20))
    ax = fig.add_subplot(1, 1, 1, projection=W)
    im, _ = imshow_norm(frame.data, ax, origin='lower',
                            interval=ZScaleInterval(),
                            stretch=LogStretch(a=1))
    plt.plot(x, y, 'xr')
    fig.colorbar(im)
    fig.savefig(path +f'STD_IN_FIELD-{filtro}.png')
    logger.info("STD's FoV has been saved as a PNG file")

    # Extract Sources

    frame_data = frame.data.byteswap().newbyteorder()

    bkg = sep.Background(frame_data)
    logger.info(f"Background estimated: {bkg.globalback} +- {bkg.globalrms}")

    kernel = np.array([[1., 2., 3., 2., 1.],
                   [2., 3., 5., 3., 2.],
                   [3., 5., 8., 5., 3.],
                   [2., 3., 5., 3., 2.],
                   [1., 2., 3., 2., 1.]])

    clean_data = frame_data - bkg
    logger.info("Background subtracted")

    objects = sep.extract(clean_data, conf["PHOTOMETRY"]["threshold"], 
                          err=bkg.globalrms, filter_kernel=kernel)
    
    logger.info("Objects catalogue in FoV has been created")

    # plot background-subtracted image
    fig = plt.figure(figsize=(15,20))
    ax = fig.add_subplot(1, 1, 1, projection=W)
    im, norm = imshow_norm(clean_data, ax, origin='lower',
                            interval=ZScaleInterval(),
                            stretch=LogStretch(a=1))

    # plot an ellipse for each object
    for i in range(len(objects)):
        e = Ellipse(xy=(objects['x'][i], objects['y'][i]),
                    width=6*objects['a'][i],
                    height=6*objects['b'][i],
                    angle=objects['theta'][i] * 180. / np.pi)
        e.set_facecolor('none')
        e.set_edgecolor('red')
        ax.add_artist(e)

    fig.savefig(path +f'SOURCES_DETECTED-{filtro}.png')
    logger.info("STD's FoV has been saved adding the objects")

    distance = np.sqrt((objects['x'] - x)**2 + (objects['y']-y)**2)

    index = np.argmin(distance)

    target = objects[index]

    X = np.array([objects['x'][index]])
    Y = np.array([objects['y'][index]])
    a = np.array([objects['a'][index]])
    b = np.array([objects['b'][index]])
    theta = np.array([objects['theta'][index]])
    flux = np.array([objects['flux'][index]])

    logger.info(f"STD name: {hd['OBJECT']}")
    logger.info(f"RA: {ra}, Dec: {dec}")
    logger.info(f"Exposure time: {t} sec")
    logger.info(f"Position: {X[0]}, {Y[0]}")
    logger.info(f"Ellipse info -> a: {a[0]}, b: {b[0]} & theta: {theta[0]}")
    logger.info(f"Flux: {flux[0]} counts")

    kronrad, krflag = sep.kron_radius(clean_data, X, Y, a, b, theta, 6.0)
    flux, fluxerr, flag = sep.sum_ellipse(clean_data, X, Y, a, b, theta, 2.5*kronrad,
                                        subpix=1)
    flag |= krflag  # combine flags into 'flag'

    logger.info(f"Kron Radius: {kronrad[0]} pxs")
    logger.info(f"AUTO FLUX: {flux[0]} counts")
    logger.info(f"AUTO ERROR FLUX: {fluxerr[0]} counts")
    logger.info(f"AUTO FLAG: {flag[0]}")

    #Estimation ZeroPoint
    mags = get_ugriz(bands_dict[target_name][2:])
    
    ZP = mags[2] + 2.5*np.log10(flux[0]/t) + (extinction_dict[filtro][0] * hd['AIRMASS'])

    dm = 0
    dflux = fluxerr[0]
    dZP = dm + dflux * (t/flux[0]) * np.log10(np.e) + (extinction_dict[filtro][1] * hd['AIRMASS'])

    #print(f"ZP value: {ZP} +- {dZP}")
    logger.info(f'Estimated ZP: {ZP} +- {dZP} for {filtro}')

    return ZP, dZP


import json
import os
import ccdproc as ccdp

def read_config(config_path):
    with open(config_path, 'r') as file:
        config = json.load(file)
    return config


def update_config(config_path, config):
    with open(config_path, 'w') as file:
        json.dump(config, file, indent=4)


def classify_images(tab):

    existence = {
        'exist_BIAS': False,
        'exist_SKYFLAT': False,
        'exist_SCIENCE': False,
        'exist_STD': False
    }

    if 'OsirisBias' in tab['OBSMODE']:
        existence['exist_BIAS'] = True
    
    if 'OsirisSkyFlat' in tab['OBSMODE']:
        existence['exist_SKYFLAT'] = True

    if 'OsirisBroadBandImage' in tab['OBSMODE']:
        existence['exist_STD'] = (len([item for item in tab['OBJECT'] if item.startswith('STD')]) != 0)
        existence['exist_SCIENCE'] = (len([item for item in tab['OBJECT'] if not item.startswith('STD')]) != 0)

    #Special case for OsirisBroadBandImage

    if ('OsirisBroadBandImage' in tab['OBSMODE']) and ('OPEN' in tab['FILTER2']):
        existence['exist_STD'] = False
        existence['exist_SCIENCE'] = True
    
    return existence


def check_files(config_path, PRG, OB):

    conf = read_config(config_path)
    directory = conf['DIRECTORIES']['PATH_DATA'] + f"{PRG}_{OB}/raw/"

    ic = ccdp.ImageFileCollection(directory, keywords=['OBSMODE','OBJECT','FILTER2','EXPTIME'])
    image_types = classify_images(ic.summary)

    # Update config based on image types
    conf['REDUCTION']['use_BIAS'] = image_types['exist_BIAS']
    conf['REDUCTION']['use_FLAT'] = image_types['exist_SKYFLAT']
    conf['REDUCTION']['use_STD'] = (image_types['exist_STD'] and image_types['exist_SKYFLAT'])
    conf['REDUCTION']['save_std'] = (image_types['exist_STD'] and image_types['exist_SKYFLAT'])
    conf['REDUCTION']['save_sky'] = image_types['exist_SCIENCE']
    conf['PHOTOMETRY']['use_photometry'] = (image_types['exist_SKYFLAT'] and image_types['exist_STD'])

    update_config(config_path, conf)


"""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.

Copyright (C) 2024 Gran Telescopio Canarias <https://www.gtc.iac.es>
Fabricio Manuel Pérez Toledo <fabricio.perez@gtc.iac.es>
"""

from setuptools import setup, find_packages

setup(
    name="sausero",
    version="0.2.0",
    packages=find_packages(where='.'),
    package_data={
        'SAUSERO    ': [
            'config/configuration.json',
            'BPM/BPM_OSIRIS_PLUS.fits'
        ],
    },
    include_package_data=True,
    install_requires=[
        "astroalign",
	"astrometry_net_client",
	"astropy",
	"astroquery",
	"ccdproc",
	"lacosmic",
	"loguru",
	"matplotlib",
	"numpy",
	"PyYAML",
	"sep"
    ],
    entry_points={
        'console_scripts': [
            'sausero=SAUSERO.OsirisDRP:run',
        ],
    },
    author="Fabricio M. Pérez-Toledo",
    author_email="fabricio.telescope@gmail.com",
    description="This software is designed to reduce Broad Band Imaging observations obtained with OSIRIS+.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/Kennicutt/SAUSERO",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
)



"""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.

Copyright (C) 2024 Gran Telescopio Canarias <https://www.gtc.iac.es>
Fabricio Manuel Pérez Toledo <fabricio.perez@gtc.iac.es>
"""

import astroalign as aa
from astropy.nddata import CCDData
from astropy.table import Table
import ccdproc as ccdp
from pathlib import Path
import time
import matplotlib.pyplot as plt
from astropy.visualization import LogStretch,imshow_norm, ZScaleInterval

from SAUSERO.Color_Codes import bcolors as bcl
from loguru import logger


class OsirisAlign:
    """This class allows the alignment of science frames (or photometry calibration frames, if applicable). 
    Afterward, we can stack them to enhance the measured flux. This step is essential for observing 
    faint sources.
    """
    
    def __init__(self, program, block, conf):
        """We initialize the class by defining important parameters.

        Args:
            program (str): Science program code
            block (str): Observational block number assigned to a science program
        """
        self.conf = conf
        self.PATH = self.conf["DIRECTORIES"]["PATH_DATA"]
        self.program = program
        self.block = block
        self.reduced_data_directory = "reduced"
        self.PATH_REDUCED = Path(self.PATH + self.program + "_" + self.block + "/" + "reduced/")
        #self.PATH_TARGET = Path(self.PATH + self.program + "_" + self.block + "/" + "object/")
        self.ic = ccdp.ImageFileCollection(self.PATH_REDUCED, keywords='*', glob_include='ADP*')


    def load_frames(self, filt, sky):
        """This method retrieves a list of science frames for each filter.

        Args:
            filt (str): Filter name

        Returns:
            list: A list of science frames for a given filter and its path
        """
        self.tab = Table(self.ic.summary)
        sub_tab = self.tab[(self.tab['filter2']==filt) & (self.tab['ssky']==sky)]
        logger.info(f"Looking for frames that have {filt} and {sky}")
        logger.info(f"{sub_tab}")
        self.total_exptime = sub_tab['exptime'].value.data[0]
        logger.info(f"Exposure time per frame: {self.total_exptime} sec")
        return self.ic.files_filtered(imgtype="SCIENCE",
                                      filtro=filt,
                                      ssky = sky,
                                      include_path=True)



    def get_each_data(self, filt, sky):
        """This methods reads a list containing the paths to several frames 
        and opens them to add them to a new list.

        Args:
            filt (str): Filter name

        Returns:
            list: List of science frames for each filter (matrices)
        """
        ccd = []
        for frame_path in self.load_frames(filt, sky=sky):
            ccd.append(CCDData.read(frame_path, unit='adu', hdu=0).data)

        return ccd



    def aligning(self, filt, sky='SKY'): #default: 30
        """This method aligns the science frames taken with the same filter.

        Args:
            filt (str): Filter name

        Returns:
            float: Stacked image obtained by combining multiple science frames.
        """
        logger.info(f"Creating cube with frames for {filt}")
        cube = self.get_each_data(filt, sky=sky)
        REF = cube[0]

        self.num=0
        logger.info(f"Number of frames in cube is: {len(cube)}")
        for IMG in cube[1:]:
            try:
                t, __ = aa.find_transform(IMG, REF, 
                                          max_control_points=self.conf["ALIGNING"]["max_control_points"]) #default: 30
                time.sleep(1)
                REF = REF.byteswap().newbyteorder()
                IMG = IMG.byteswap().newbyteorder()
                align, footprint = aa.apply_transform(t, IMG, REF)
                REF = REF + align
                self.num += 1
                logger.info(f"Image NO: {self.num}/{len(cube)}")
            except aa.MaxIterError:
                #print(aa.MaxIterError)
                logger.error(f"{bcl.FAIL}ERROR{bcl.ENDC}: {aa.MaxIterError}")
                pass
            except ValueError:
                #print(ValueError)
                logger.error(f"{bcl.FAIL}ERROR{bcl.ENDC}: {ValueError}")
                pass
            except TypeError:
                #print(TypeError)
                logger.error(f"{bcl.FAIL}ERROR{bcl.ENDC}: {TypeError}")
                pass

        return REF


def show_picture(cube, a=1):
    """This method allows us to show the image.

    Args:
        cube (float): Stacked image.
        a (int, optional): Factor applied. Defaults: 1.
    """
    fig = plt.figure(figsize=(15,20))
    ax = fig.add_subplot(1, 1, 1)
    im, norm = imshow_norm(cube, ax, origin='lower',
                           interval=ZScaleInterval(),
                           stretch=LogStretch(a=a))
    fig.colorbar(im)
    plt.show()
    
def save_fits(image, header, wcs, fname):
    """This method saves the stacked image.

    Args:
        image (float): Data to save.
        header (str): Header for the stacked image.
        wcs (str): WCS information for the stacked image.
        fname (str): Name for the stacked image.
    """
    ccd = CCDData(data=image, header=header, wcs=wcs, unit='adu')
    ccd.write(fname, overwrite=True)
    logger.info(f"New image has been created: {fname}")